@@ -6,9 +6,9 @@
 */
 
 #include "rsync.h"
+#include "zlib/zlib.h"
 #include <time.h>
 
-extern int am_sender;
 extern int eol_nulls;
 extern int recurse;
 extern int xfer_dirs;
@@ -19,11 +19,14 @@ extern int preserve_uid;
 extern int preserve_gid;
 extern int always_checksum;
 extern int do_compression;
+extern int def_compress_level;
 extern int protocol_version;
 extern char *batch_name;
 
 extern struct filter_list_struct filter_list;
 
+static int tweaked_compress_level;
+
 static int *flag_ptr[] = {
 	&recurse,		/* 0 */
 	&preserve_uid,		/* 1 */
@@ -33,7 +36,7 @@ static int *flag_ptr[] = {
 	&preserve_hard_links,	/* 5 */
 	&always_checksum,	/* 6 */
 	&xfer_dirs,		/* 7 (protocol 29) */
-	&do_compression,	/* 8 (protocol 29) */
+	&tweaked_compress_level,/* 8 (protocol 29) */
 	NULL
 };
 
@@ -54,6 +57,12 @@ void write_stream_flags(int fd)
 {
 	int i, flags;
 
+#if Z_DEFAULT_COMPRESSION == -1
+	tweaked_compress_level = do_compression ? def_compress_level + 2 : 0;
+#else
+#error internal logic error!  Fix def_compress_level logic above and below too!
+#endif
+
 	/* Start the batch file with a bitmap of data-stream-affecting
 	 * flags. */
 	if (protocol_version < 29)
@@ -88,6 +97,13 @@ void read_stream_flags(int fd)
 		else if (xfer_dirs < 2)
 			xfer_dirs = 0;
 	}
+
+	if (tweaked_compress_level == 0 || tweaked_compress_level == 2)
+		do_compression = 0;
+	else {
+		do_compression = 1;
+		def_compress_level = tweaked_compress_level - 2;
+	}
 }
 
 static void write_arg(int fd, char *arg)
@@ -139,7 +155,7 @@ static void write_filter_rules(int fd)
  * (hopefully) work. */
 void write_batch_shell_file(int argc, char *argv[], int file_arg_cnt)
 {
-	int fd, i;
+	int fd, i, len;
 	char *p, filename[MAXPATHLEN];
 
 	stringjoin(filename, sizeof filename,
@@ -148,7 +164,7 @@ void write_batch_shell_file(int argc, char *argv[], int file_arg_cnt)
 		     S_IRUSR | S_IWUSR | S_IEXEC);
 	if (fd < 0) {
 		rsyserr(FERROR, errno, "Batch file %s open error",
-			safe_fname(filename));
+			filename);
 		exit_cleanup(1);
 	}
 
@@ -175,11 +191,12 @@ void write_batch_shell_file(int argc, char *argv[], int file_arg_cnt)
 			continue;
 		}
 		write(fd, " ", 1);
-		if (strncmp(p, "--write-batch", 13) == 0) {
+		if (strncmp(p, "--write-batch", len = 13) == 0
+		 || strncmp(p, "--only-write-batch", len = 18) == 0) {
 			write(fd, "--read-batch", 12);
-			if (p[13] == '=') {
+			if (p[len] == '=') {
 				write(fd, "=", 1);
-				write_arg(fd, p + 14);
+				write_arg(fd, p + len + 1);
 			}
 		} else
 			write_arg(fd, p);
@@ -193,42 +210,7 @@ void write_batch_shell_file(int argc, char *argv[], int file_arg_cnt)
 		write_filter_rules(fd);
 	if (write(fd, "\n", 1) != 1 || close(fd) < 0) {
 		rsyserr(FERROR, errno, "Batch file %s write error",
-			safe_fname(filename));
+			filename);
 		exit_cleanup(1);
 	}
 }
-
-void show_flist(int index, struct file_struct **fptr)
-{
-	/*  for debugging    show_flist(flist->count, flist->files * */
-
-	int i;
-	for (i = 0; i < index; i++) {
-		rprintf(FINFO, "flist->flags=%#x\n", fptr[i]->flags);
-		rprintf(FINFO, "flist->modtime=%#lx\n",
-			(long unsigned) fptr[i]->modtime);
-		rprintf(FINFO, "flist->length=%.0f\n",
-			(double) fptr[i]->length);
-		rprintf(FINFO, "flist->mode=%#o\n", (int) fptr[i]->mode);
-		rprintf(FINFO, "flist->basename=%s\n",
-			safe_fname(fptr[i]->basename));
-		if (fptr[i]->dirname) {
-			rprintf(FINFO, "flist->dirname=%s\n",
-				safe_fname(fptr[i]->dirname));
-		}
-		if (am_sender && fptr[i]->dir.root) {
-			rprintf(FINFO, "flist->dir.root=%s\n",
-				safe_fname(fptr[i]->dir.root));
-		}
-	}
-}
-
-/* for debugging */
-void show_argvs(int argc, char *argv[])
-{
-	int i;
-
-	rprintf(FINFO, "BATCH.C:show_argvs,argc=%d\n", argc);
-	for (i = 0; i < argc; i++)
-		rprintf(FINFO, "i=%d,argv[i]=%s\n", i, safe_fname(argv[i]));
-}
