@@ -34,8 +34,6 @@ extern int list_only;
 extern int recurse;
 extern int io_error;
 extern int local_server;
-extern int saw_delete_opt;
-extern int saw_delete_excluded_opt;
 extern int prune_empty_dirs;
 extern int delete_mode;
 extern int delete_excluded;
@@ -562,7 +560,7 @@ static int rule_matches(char *name, struct filter_struct *ex, int name_is_dir)
 		if (litmatch_array(pattern, strings, slash_handling))
 			return ret_match;
 	} else if (anchored_match) {
-		if (strcmp(name,pattern) == 0)
+		if (strcmp(strings[0], pattern) == 0)
 			return ret_match;
 	} else {
 		int l1 = strlen(name);
@@ -1185,8 +1183,8 @@ void recv_filter_list(int f_in)
 	char line[BIGPATHBUFLEN];
 	int xflags = protocol_version >= 29 ? 0 : XFLG_OLD_PREFIXES;
 	int receiver_wants_list = prune_empty_dirs
-	    || (saw_delete_opt
-	     && (!saw_delete_excluded_opt || protocol_version >= 29));
+	    || (delete_mode
+	     && (!delete_excluded || protocol_version >= 29));
 	unsigned int len;
 
 	if (!local_server && (am_sender || receiver_wants_list)) {
