@@ -20,13 +20,18 @@
 
 #include "rsync.h"
 
-extern int dry_run;
 extern int verbose;
+extern int link_dest;
 extern int make_backups;
+extern int log_format_has_i;
+extern char *basis_dir[];
 extern struct file_list *the_file_list;
 
 #ifdef SUPPORT_HARD_LINKS
 
+#define SKIPPED_LINK (-1)
+#define FINISHED_LINK (-2)
+
 #define FPTR(i) (the_file_list->files[i])
 #define LINKED(p1,p2) (FPTR(p1)->F_DEV == FPTR(p2)->F_DEV \
 		    && FPTR(p1)->F_INODE == FPTR(p2)->F_INODE)
@@ -52,7 +57,7 @@ static int hlink_count;
  * linked, and replace the dev+inode data with the hlindex+next linked list. */
 static void link_idev_data(void)
 {
-	int head, from, to, start;
+	int cur, from, to, start;
 
 	alloc_pool_t hlink_pool;
 	alloc_pool_t idev_pool = the_file_list->hlink_pool;
@@ -62,31 +67,31 @@ static void link_idev_data(void)
 
 	for (from = to = 0; from < hlink_count; from++) {
 		start = from;
-		head = hlink_list[start];
-		while (from < hlink_count-1
-		    && LINKED(hlink_list[from], hlink_list[from+1])) {
-			pool_free(idev_pool, 0, FPTR(hlink_list[from])->link_u.idev);
-			FPTR(hlink_list[from])->link_u.links = pool_talloc(hlink_pool,
+		while (1) {
+			cur = hlink_list[from];
+			if (from == hlink_count-1
+			    || !LINKED(cur, hlink_list[from+1]))
+				break;
+			pool_free(idev_pool, 0, FPTR(cur)->link_u.idev);
+			FPTR(cur)->link_u.links = pool_talloc(hlink_pool,
 			    struct hlink, 1, "hlink_list");
 
-			FPTR(hlink_list[from])->F_HLINDEX = to;
-			FPTR(hlink_list[from])->F_NEXT = hlink_list[from+1];
-			from++;
+			FPTR(cur)->F_HLINDEX = to;
+			FPTR(cur)->F_NEXT = hlink_list[++from];
 		}
+		pool_free(idev_pool, 0, FPTR(cur)->link_u.idev);
 		if (from > start) {
-			pool_free(idev_pool, 0, FPTR(hlink_list[from])->link_u.idev);
-			FPTR(hlink_list[from])->link_u.links = pool_talloc(hlink_pool,
+			int head = hlink_list[start];
+			FPTR(cur)->link_u.links = pool_talloc(hlink_pool,
 			    struct hlink, 1, "hlink_list");
 
 			FPTR(head)->flags |= FLAG_HLINK_TOL;
-			FPTR(hlink_list[from])->F_HLINDEX = to;
-			FPTR(hlink_list[from])->F_NEXT = head;
-			FPTR(hlink_list[from])->flags |= FLAG_HLINK_EOL;
+			FPTR(cur)->F_HLINDEX = to;
+			FPTR(cur)->F_NEXT = head;
+			FPTR(cur)->flags |= FLAG_HLINK_EOL;
 			hlink_list[to++] = head;
-		} else {
-			pool_free(idev_pool, 0, FPTR(head)->link_u.idev);
-			FPTR(head)->link_u.idev = NULL;
-		}
+		} else
+			FPTR(cur)->link_u.links = NULL;
 	}
 
 	if (!to) {
@@ -110,9 +115,6 @@ void init_hard_links(void)
 #ifdef SUPPORT_HARD_LINKS
 	int i;
 
-	if (the_file_list->count < 2)
-		return;
-
 	if (hlink_list)
 		free(hlink_list);
 
@@ -136,18 +138,90 @@ void init_hard_links(void)
 #endif
 }
 
-int hard_link_check(struct file_struct *file, int ndx, int skip)
+#ifdef SUPPORT_HARD_LINKS
+static int maybe_hard_link(struct file_struct *file, int ndx,
+			   char *fname, int statret, STRUCT_STAT *st,
+			   char *toname, STRUCT_STAT *to_st,
+			   int itemizing, enum logcode code)
+{
+	if (statret == 0) {
+		if (st->st_dev == to_st->st_dev
+		 && st->st_ino == to_st->st_ino) {
+			if (itemizing) {
+				itemize(file, ndx, statret, st,
+					ITEM_LOCAL_CHANGE | ITEM_XNAME_FOLLOWS,
+					0, "");
+			}
+			return 0;
+		}
+		if (make_backups) {
+			if (!make_backup(fname))
+				return -1;
+		} else if (robust_unlink(fname)) {
+			rsyserr(FERROR, errno, "unlink %s failed",
+				full_fname(fname));
+			return -1;
+		}
+	}
+	return hard_link_one(file, ndx, fname, statret, st, toname,
+			     0, itemizing, code);
+}
+#endif
+
+int hard_link_check(struct file_struct *file, int ndx, char *fname,
+		    int statret, STRUCT_STAT *st, int itemizing,
+		    enum logcode code, int skip)
 {
 #ifdef SUPPORT_HARD_LINKS
-	if (!hlink_list || !file->link_u.links)
-		return 0;
+	int head;
 	if (skip && !(file->flags & FLAG_HLINK_EOL))
-		hlink_list[file->F_HLINDEX] = file->F_NEXT;
-	if (hlink_list[file->F_HLINDEX] != ndx) {
-		if (verbose > 2) {
+		head = hlink_list[file->F_HLINDEX] = file->F_NEXT;
+	else
+		head = hlink_list[file->F_HLINDEX];
+	if (ndx != head) {
+		struct file_struct *head_file = FPTR(head);
+		if (!log_format_has_i && verbose > 1) {
 			rprintf(FINFO, "\"%s\" is a hard link\n",
-				safe_fname(f_name(file)));
+				f_name(file, NULL));
 		}
+		if (head_file->F_HLINDEX == FINISHED_LINK) {
+			STRUCT_STAT st2, st3;
+			char *toname = f_name(head_file, NULL);
+			if (link_stat(toname, &st2, 0) < 0) {
+				rsyserr(FERROR, errno, "stat %s failed",
+					full_fname(toname));
+				return -1;
+			}
+			if (statret < 0 && basis_dir[0] != NULL) {
+				char cmpbuf[MAXPATHLEN];
+				int j = 0;
+				do {
+					pathjoin(cmpbuf, MAXPATHLEN, basis_dir[j], fname);
+					if (link_stat(cmpbuf, &st3, 0) < 0)
+						continue;
+					if (link_dest) {
+						if (st2.st_dev != st3.st_dev
+						 || st2.st_ino != st3.st_ino)
+							continue;
+						statret = 1;
+						st = &st3;
+						if (verbose < 2 || !log_format_has_i)
+							itemizing = code = 0;
+						break;
+					}
+					if (!unchanged_file(cmpbuf, file, &st3))
+						continue;
+					statret = 1;
+					st = &st3;
+					if (unchanged_attrs(file, &st3))
+						break;
+				} while (basis_dir[++j] != NULL);
+			}
+			maybe_hard_link(file, ndx, fname, statret, st,
+					toname, &st2, itemizing, code);
+			file->F_HLINDEX = FINISHED_LINK;
+		} else
+			file->F_HLINDEX = SKIPPED_LINK;
 		return 1;
 	}
 #endif
@@ -160,10 +234,14 @@ int hard_link_one(struct file_struct *file, int ndx, char *fname,
 		  int itemizing, enum logcode code)
 {
 	if (do_link(toname, fname)) {
-		if (verbose) {
-			rsyserr(FERROR, errno, "link %s => %s failed",
-				full_fname(fname), safe_fname(toname));
-		}
+		if (terse) {
+			if (!verbose)
+				return -1;
+			code = FINFO;
+		} else
+			code = FERROR;
+		rsyserr(code, errno, "link %s => %s failed",
+			full_fname(fname), toname);
 		return -1;
 	}
 
@@ -172,10 +250,8 @@ int hard_link_one(struct file_struct *file, int ndx, char *fname,
 			ITEM_LOCAL_CHANGE | ITEM_XNAME_FOLLOWS, 0,
 			terse ? "" : toname);
 	}
-	if (code && verbose && !terse) {
-		rprintf(code, "%s => %s\n",
-			safe_fname(fname), safe_fname(toname));
-	}
+	if (code && verbose && !terse)
+		rprintf(code, "%s => %s\n", fname, toname);
 	return 0;
 }
 #endif
@@ -190,7 +266,8 @@ void hard_link_cluster(struct file_struct *file, int master, int itemizing,
 	STRUCT_STAT st1, st2;
 	int statret, ndx = master;
 
-	if (link_stat(f_name_to(file, hlink1), &st1, 0) < 0)
+	file->F_HLINDEX = FINISHED_LINK;
+	if (link_stat(f_name(file, hlink1), &st1, 0) < 0)
 		return;
 	if (!(file->flags & FLAG_HLINK_TOL)) {
 		while (!(file->flags & FLAG_HLINK_EOL)) {
@@ -201,33 +278,13 @@ void hard_link_cluster(struct file_struct *file, int master, int itemizing,
 	do {
 		ndx = file->F_NEXT;
 		file = FPTR(ndx);
-		if (ndx == master)
+		if (file->F_HLINDEX != SKIPPED_LINK)
 			continue;
-		hlink2 = f_name(file);
-		if ((statret = link_stat(hlink2, &st2, 0)) == 0) {
-			if (st2.st_dev == st1.st_dev
-			    && st2.st_ino == st1.st_ino) {
-				if (itemizing) {
-					itemize(file, ndx, statret, &st2,
-						ITEM_LOCAL_CHANGE | ITEM_XNAME_FOLLOWS,
-						0, "");
-				}
-				continue;
-			}
-			if (make_backups) {
-				if (!make_backup(hlink2))
-					continue;
-			} else if (robust_unlink(hlink2)) {
-				if (verbose > 0) {
-					rsyserr(FINFO, errno,
-						"unlink %s failed",
-						full_fname(hlink2));
-				}
-				continue;
-			}
-		}
-		hard_link_one(file, ndx, hlink2, statret,
-			      &st2, hlink1, 0, itemizing, code);
+		hlink2 = f_name(file, NULL);
+		statret = link_stat(hlink2, &st2, 0);
+		maybe_hard_link(file, ndx, hlink2, statret, &st2,
+				hlink1, &st1, itemizing, code);
+		file->F_HLINDEX = FINISHED_LINK;
 	} while (!(file->flags & FLAG_HLINK_EOL));
 #endif
 }
