@@ -769,11 +769,11 @@ void log_exit(int code, const char *file, int line)
 
 		/* VANISHED is not an error, only a warning */
 		if (code == RERR_VANISHED) {
-			rprintf(FINFO, "rsync warning: %s (code %d) at %s(%d) [%s]\n", 
-				name, code, file, line, who_am_i());
+			rprintf(FINFO, "rsync warning: %s (code %d) at %s(%d) [%s=%s]\n", 
+				name, code, file, line, who_am_i(), RSYNC_VERSION);
 		} else {
-			rprintf(FERROR, "rsync error: %s (code %d) at %s(%d) [%s]\n",
-				name, code, file, line, who_am_i());
+			rprintf(FERROR, "rsync error: %s (code %d) at %s(%d) [%s=%s]\n",
+				name, code, file, line, who_am_i(), RSYNC_VERSION);
 		}
 	}
 }
