@@ -20,69 +20,47 @@
 #include "rsync.h"
 
 extern int verbose;
-extern int am_server;
 extern int do_progress;
 extern int checksum_seed;
-extern int inplace;
-extern int make_backups;
+extern int append_mode;
 
-typedef unsigned short tag;
-
-#define TABLESIZE (1<<16)
-#define NULL_TAG ((size_t)-1)
+int updating_basis_file;
 
 static int false_alarms;
-static int tag_hits;
+static int hash_hits;
 static int matches;
 static int64 data_transfer;
 
 static int total_false_alarms;
-static int total_tag_hits;
+static int total_hash_hits;
 static int total_matches;
 
 extern struct stats stats;
 
-struct target {
-	tag t;
-	size_t i;
-};
+#define TABLESIZE (1<<16)
 
-static struct target *targets;
-
-static size_t *tag_table;
-
-#define gettag2(s1,s2) (((s1) + (s2)) & 0xFFFF)
-#define gettag(sum) gettag2((sum)&0xFFFF,(sum)>>16)
-
-static int compare_targets(struct target *t1,struct target *t2)
-{
-	return (int)t1->t - (int)t2->t;
-}
+static int32 *hash_table;
 
+#define SUM2HASH2(s1,s2) (((s1) + (s2)) & 0xFFFF)
+#define SUM2HASH(sum) SUM2HASH2((sum)&0xFFFF,(sum)>>16)
 
 static void build_hash_table(struct sum_struct *s)
 {
-	size_t i;
+	int32 i;
 
-	if (!tag_table)
-		tag_table = new_array(size_t, TABLESIZE);
-
-	targets = new_array(struct target, s->count);
-	if (!tag_table || !targets)
-		out_of_memory("build_hash_table");
-
-	for (i = 0; i < s->count; i++) {
-		targets[i].i = i;
-		targets[i].t = gettag(s->sums[i].sum1);
+	if (!hash_table) {
+		hash_table = new_array(int32, TABLESIZE);
+		if (!hash_table)
+			out_of_memory("build_hash_table");
 	}
 
-	qsort(targets,s->count,sizeof(targets[0]),(int (*)())compare_targets);
+	memset(hash_table, 0xFF, TABLESIZE * sizeof hash_table[0]);
 
-	for (i = 0; i < TABLESIZE; i++)
-		tag_table[i] = NULL_TAG;
-
-	for (i = s->count; i-- > 0; )
-		tag_table[targets[i].t] = i;
+	for (i = 0; i < s->count; i++) {
+		uint32 t = SUM2HASH(s->sums[i].sum1);
+		s->sums[i].chain = hash_table[t];
+		hash_table[t] = i;
+	}
 }
 
 
@@ -101,17 +79,20 @@ static OFF_T last_match;
  * @param i If >0, the number of a matched token.  If 0, indicates we
  * have only literal data.
  **/
-static void matched(int f,struct sum_struct *s,struct map_struct *buf,
-		    OFF_T offset,int i)
+static void matched(int f, struct sum_struct *s, struct map_struct *buf,
+		    OFF_T offset, int32 i)
 {
-	OFF_T n = offset - last_match;
-	OFF_T j;
+	int32 n = offset - last_match; /* max value: block_size (int32) */
+	int32 j;
 
-	if (verbose > 2 && i >= 0)
-		rprintf(FINFO,"match at %.0f last_match=%.0f j=%d len=%u n=%.0f\n",
-			(double)offset,(double)last_match,i,s->sums[i].len,(double)n);
+	if (verbose > 2 && i >= 0) {
+		rprintf(FINFO,
+			"match at %.0f last_match=%.0f j=%d len=%ld n=%ld\n",
+			(double)offset, (double)last_match, i,
+			(long)s->sums[i].len, (long)n);
+	}
 
-	send_token(f,i,buf,last_match,n,i<0?0:s->sums[i].len);
+	send_token(f, i, buf, last_match, n, i < 0 ? 0 : s->sums[i].len);
 	data_transfer += n;
 
 	if (i >= 0) {
@@ -120,22 +101,17 @@ static void matched(int f,struct sum_struct *s,struct map_struct *buf,
 	}
 
 	for (j = 0; j < n; j += CHUNK_SIZE) {
-		int n1 = MIN(CHUNK_SIZE,n-j);
-		sum_update(map_ptr(buf,last_match+j,n1),n1);
+		int32 n1 = MIN(CHUNK_SIZE, n - j);
+		sum_update(map_ptr(buf, last_match + j, n1), n1);
 	}
 
-
 	if (i >= 0)
 		last_match = offset + s->sums[i].len;
 	else
 		last_match = offset;
 
-	if (buf && do_progress) {
+	if (buf && do_progress)
 		show_progress(last_match, buf->file_size);
-
-		if (i == -1)
-			end_progress(buf->file_size);
-	}
 }
 
 
@@ -143,8 +119,7 @@ static void hash_search(int f,struct sum_struct *s,
 			struct map_struct *buf, OFF_T len)
 {
 	OFF_T offset, end, backup;
-	unsigned int k;
-	size_t want_i;
+	int32 k, want_i;
 	char sum2[SUM_LENGTH];
 	uint32 s1, s2, sum;
 	int more;
@@ -155,11 +130,11 @@ static void hash_search(int f,struct sum_struct *s,
 	want_i = 0;
 
 	if (verbose > 2) {
-		rprintf(FINFO,"hash search b=%u len=%.0f\n",
-			s->blength, (double)len);
+		rprintf(FINFO, "hash search b=%ld len=%.0f\n",
+			(long)s->blength, (double)len);
 	}
 
-	k = MIN(len, s->blength);
+	k = (int32)MIN(len, (OFF_T)s->blength);
 
 	map = (schar *)map_ptr(buf, 0, k);
 
@@ -167,51 +142,54 @@ static void hash_search(int f,struct sum_struct *s,
 	s1 = sum & 0xFFFF;
 	s2 = sum >> 16;
 	if (verbose > 3)
-		rprintf(FINFO, "sum=%.8x k=%u\n", sum, k);
+		rprintf(FINFO, "sum=%.8x k=%ld\n", sum, (long)k);
 
 	offset = 0;
 
 	end = len + 1 - s->sums[s->count-1].len;
 
 	if (verbose > 3) {
-		rprintf(FINFO, "hash search s->blength=%u len=%.0f count=%.0f\n",
-			s->blength, (double)len, (double)s->count);
+		rprintf(FINFO, "hash search s->blength=%ld len=%.0f count=%.0f\n",
+			(long)s->blength, (double)len, (double)s->count);
 	}
 
 	do {
-		tag t = gettag2(s1,s2);
 		int done_csum2 = 0;
-		size_t j = tag_table[t];
+		int32 i;
 
-		if (verbose > 4)
-			rprintf(FINFO,"offset=%.0f sum=%08x\n",(double)offset,sum);
+		if (verbose > 4) {
+			rprintf(FINFO, "offset=%.0f sum=%04x%04x\n",
+				(double)offset, s2 & 0xFFFF, s1 & 0xFFFF);
+		}
 
-		if (j == NULL_TAG)
-			goto null_tag;
+		i = hash_table[SUM2HASH2(s1,s2)];
+		if (i < 0)
+			goto null_hash;
 
 		sum = (s1 & 0xffff) | (s2 << 16);
-		tag_hits++;
+		hash_hits++;
 		do {
-			unsigned int l;
-			size_t i = targets[j].i;
+			int32 l;
 
 			if (sum != s->sums[i].sum1)
 				continue;
 
 			/* also make sure the two blocks are the same length */
-			l = MIN((OFF_T)s->blength, len-offset);
+			l = (int32)MIN((OFF_T)s->blength, len-offset);
 			if (l != s->sums[i].len)
 				continue;
 
-			/* inplace: ensure chunk's offset is either >= our
+			/* in-place: ensure chunk's offset is either >= our
 			 * offset or that the data didn't move. */
-			if (inplace && !make_backups && s->sums[i].offset < offset
+			if (updating_basis_file && s->sums[i].offset < offset
 			    && !(s->sums[i].flags & SUMFLG_SAME_OFFSET))
 				continue;
 
-			if (verbose > 3)
-				rprintf(FINFO,"potential match at %.0f target=%.0f %.0f sum=%08x\n",
-					(double)offset,(double)j,(double)i,sum);
+			if (verbose > 3) {
+				rprintf(FINFO,
+					"potential match at %.0f i=%ld sum=%08x\n",
+					(double)offset, (long)i, sum);
+			}
 
 			if (!done_csum2) {
 				map = (schar *)map_ptr(buf,offset,l);
@@ -224,12 +202,12 @@ static void hash_search(int f,struct sum_struct *s,
 				continue;
 			}
 
-			/* If inplace is enabled, the best possible match is
+			/* When updating in-place, the best possible match is
 			 * one with an identical offset, so we prefer that over
 			 * the following want_i optimization. */
-			if (inplace && !make_backups) {
-				do {
-					size_t i2 = targets[j].i;
+			if (updating_basis_file) {
+				int32 i2;
+				for (i2 = i; i2 >= 0; i2 = s->sums[i2].chain) {
 					if (s->sums[i2].offset != offset)
 						continue;
 					if (i2 != i) {
@@ -244,13 +222,13 @@ static void hash_search(int f,struct sum_struct *s,
 					 * both the sender and the receiver. */
 					s->sums[i].flags |= SUMFLG_SAME_OFFSET;
 					goto set_want_i;
-				} while (++j < s->count && targets[j].t == t);
+				}
 			}
 
 			/* we've found a match, but now check to see
 			 * if want_i can hint at a better match. */
 			if (i != want_i && want_i < s->count
-			    && (!inplace || make_backups || s->sums[want_i].offset >= offset
+			    && (!updating_basis_file || s->sums[want_i].offset >= offset
 			     || s->sums[want_i].flags & SUMFLG_SAME_OFFSET)
 			    && sum == s->sums[want_i].sum1
 			    && memcmp(sum2, s->sums[want_i].sum2, s->s2length) == 0) {
@@ -263,16 +241,16 @@ static void hash_search(int f,struct sum_struct *s,
 
 			matched(f,s,buf,offset,i);
 			offset += s->sums[i].len - 1;
-			k = MIN(s->blength, len-offset);
+			k = (int32)MIN((OFF_T)s->blength, len-offset);
 			map = (schar *)map_ptr(buf, offset, k);
 			sum = get_checksum1((char *)map, k);
 			s1 = sum & 0xFFFF;
 			s2 = sum >> 16;
 			matches++;
 			break;
-		} while (++j < s->count && targets[j].t == t);
+		} while ((i = s->sums[i].chain) >= 0);
 
-	null_tag:
+	  null_hash:
 		backup = offset - last_match;
 		/* We sometimes read 1 byte prior to last_match... */
 		if (backup < 0)
@@ -298,14 +276,12 @@ static void hash_search(int f,struct sum_struct *s,
 		   match. The 3 reads are caused by the
 		   running match, the checksum update and the
 		   literal send. */
-		if (backup >= CHUNK_SIZE + s->blength
-		    && end - offset > CHUNK_SIZE) {
-			matched(f,s,buf,offset - s->blength, -2);
-		}
+		if (backup >= s->blength+CHUNK_SIZE && end-offset > CHUNK_SIZE)
+			matched(f, s, buf, offset - s->blength, -2);
 	} while (++offset < end);
 
-	matched(f,s,buf,len,-1);
-	map_ptr(buf,len-1,1);
+	matched(f, s, buf, len, -1);
+	map_ptr(buf, len-1, 1);
 }
 
 
@@ -329,13 +305,32 @@ void match_sums(int f, struct sum_struct *s, struct map_struct *buf, OFF_T len)
 
 	last_match = 0;
 	false_alarms = 0;
-	tag_hits = 0;
+	hash_hits = 0;
 	matches = 0;
 	data_transfer = 0;
 
 	sum_init(checksum_seed);
 
-	if (len > 0 && s->count>0) {
+	if (append_mode) {
+		OFF_T j = 0;
+		for (j = CHUNK_SIZE; j < s->flength; j += CHUNK_SIZE) {
+			if (buf && do_progress)
+				show_progress(last_match, buf->file_size);
+			sum_update(map_ptr(buf, last_match, CHUNK_SIZE),
+				   CHUNK_SIZE);
+			last_match = j;
+		}
+		if (last_match < s->flength) {
+			int32 len = s->flength - last_match;
+			if (buf && do_progress)
+				show_progress(last_match, buf->file_size);
+			sum_update(map_ptr(buf, last_match, len), len);
+			last_match = s->flength;
+		}
+		s->count = 0;
+	}
+
+	if (len > 0 && s->count > 0) {
 		build_hash_table(s);
 
 		if (verbose > 2)
@@ -348,11 +343,9 @@ void match_sums(int f, struct sum_struct *s, struct map_struct *buf, OFF_T len)
 	} else {
 		OFF_T j;
 		/* by doing this in pieces we avoid too many seeks */
-		for (j = 0; j < len-CHUNK_SIZE; j += CHUNK_SIZE) {
-			int n1 = MIN(CHUNK_SIZE,(len-CHUNK_SIZE)-j);
-			matched(f,s,buf,j+n1,-2);
-		}
-		matched(f,s,buf,len,-1);
+		for (j = last_match + CHUNK_SIZE; j < len; j += CHUNK_SIZE)
+			matched(f, s, buf, j, -2);
+		matched(f, s, buf, len, -1);
 	}
 
 	sum_end(file_sum);
@@ -364,16 +357,11 @@ void match_sums(int f, struct sum_struct *s, struct map_struct *buf, OFF_T len)
 		rprintf(FINFO,"sending file_sum\n");
 	write_buf(f,file_sum,MD4_SUM_LENGTH);
 
-	if (targets) {
-		free(targets);
-		targets=NULL;
-	}
-
 	if (verbose > 2)
-		rprintf(FINFO, "false_alarms=%d tag_hits=%d matches=%d\n",
-			false_alarms, tag_hits, matches);
+		rprintf(FINFO, "false_alarms=%d hash_hits=%d matches=%d\n",
+			false_alarms, hash_hits, matches);
 
-	total_tag_hits += tag_hits;
+	total_hash_hits += hash_hits;
 	total_false_alarms += false_alarms;
 	total_matches += matches;
 	stats.literal_data += data_transfer;
@@ -385,8 +373,7 @@ void match_report(void)
 		return;
 
 	rprintf(FINFO,
-		"total: matches=%d  tag_hits=%d  false_alarms=%d data=%.0f\n",
-		total_matches,total_tag_hits,
-		total_false_alarms,
+		"total: matches=%d  hash_hits=%d  false_alarms=%d data=%.0f\n",
+		total_matches, total_hash_hits, total_false_alarms,
 		(double)stats.literal_data);
 }
