@@ -491,8 +491,8 @@ static FILE *OpenConfFile( char *FileName )
   OpenedFile = fopen( FileName, "r" );
   if( NULL == OpenedFile )
     {
-    rsyserr(FERROR, errno, "rsync: unable to open configuration file \"%s\"",
-	    safe_fname(FileName));
+    rsyserr(FERROR, errno, "unable to open configuration file \"%s\"",
+	    FileName);
     }
 
   return( OpenedFile );
