@@ -21,7 +21,7 @@
 #include "rsync.h"
 
 extern int verbose;
-extern int dry_run;
+extern int do_xfers;
 extern int am_daemon;
 extern int am_server;
 extern int do_progress;
@@ -30,20 +30,18 @@ extern int log_format_has_i;
 extern int daemon_log_format_has_i;
 extern int csum_length;
 extern int read_batch;
+extern int write_batch;
 extern int batch_gen_fd;
 extern int protocol_version;
 extern int relative_paths;
-extern int keep_dirlinks;
 extern int preserve_hard_links;
 extern int preserve_perms;
-extern int io_error;
 extern int basis_dir_cnt;
 extern int make_backups;
 extern int cleanup_got_literal;
 extern int remove_sent_files;
-extern int module_id;
-extern int ignore_errors;
-extern int orig_umask;
+extern int append_mode;
+extern int sparse_files;
 extern int keep_partial;
 extern int checksum_seed;
 extern int inplace;
@@ -56,70 +54,9 @@ extern char *basis_dir[];
 extern struct file_list *the_file_list;
 extern struct filter_list_struct server_filter_list;
 
-#define SLOT_SIZE	(16*1024)	/* Desired size in bytes */
-#define PER_SLOT_BITS	(SLOT_SIZE * 8) /* Number of bits per slot */
-#define PER_SLOT_INTS	(SLOT_SIZE / 4) /* Number of int32s per slot */
-
-static uint32 **delayed_bits = NULL;
-static int delayed_slot_cnt = 0;
+static struct bitbag *delayed_bits = NULL;
 static int phase = 0;
 
-static void init_delayed_bits(int max_ndx)
-{
-	delayed_slot_cnt = (max_ndx + PER_SLOT_BITS - 1) / PER_SLOT_BITS;
-
-	if (!(delayed_bits = (uint32**)calloc(delayed_slot_cnt, sizeof (uint32*))))
-		out_of_memory("set_delayed_bit");
-}
-
-static void set_delayed_bit(int ndx)
-{
-	int slot = ndx / PER_SLOT_BITS;
-	ndx %= PER_SLOT_BITS;
-
-	if (!delayed_bits[slot]) {
-		if (!(delayed_bits[slot] = (uint32*)calloc(PER_SLOT_INTS, 4)))
-			out_of_memory("set_delayed_bit");
-	}
-
-	delayed_bits[slot][ndx/32] |= 1u << (ndx % 32);
-}
-
-/* Call this with -1 to start checking from 0.  Returns -1 at the end. */
-static int next_delayed_bit(int after)
-{
-	uint32 bits, mask;
-	int i, ndx = after + 1;
-	int slot = ndx / PER_SLOT_BITS;
-	ndx %= PER_SLOT_BITS;
-
-	mask = (1u << (ndx % 32)) - 1;
-	for (i = ndx / 32; slot < delayed_slot_cnt; slot++, i = mask = 0) {
-		if (!delayed_bits[slot])
-			continue;
-		for ( ; i < PER_SLOT_INTS; i++, mask = 0) {
-			if (!(bits = delayed_bits[slot][i] & ~mask))
-				continue;
-			/* The xor magic figures out the lowest enabled bit in
-			 * bits, and the switch quickly computes log2(bit). */
-			switch (bits ^ (bits & (bits-1))) {
-#define LOG2(n) case 1u << n: return slot*PER_SLOT_BITS + i*32 + n
-			    LOG2(0);  LOG2(1);  LOG2(2);  LOG2(3);
-			    LOG2(4);  LOG2(5);  LOG2(6);  LOG2(7);
-			    LOG2(8);  LOG2(9);  LOG2(10); LOG2(11);
-			    LOG2(12); LOG2(13); LOG2(14); LOG2(15);
-			    LOG2(16); LOG2(17); LOG2(18); LOG2(19);
-			    LOG2(20); LOG2(21); LOG2(22); LOG2(23);
-			    LOG2(24); LOG2(25); LOG2(26); LOG2(27);
-			    LOG2(28); LOG2(29); LOG2(30); LOG2(31);
-			}
-			return -1; /* impossible... */
-		}
-	}
-
-	return -1;
-}
-
 
 /*
  * get_tmpname() - create a tmp filename for a given filename
@@ -170,8 +107,7 @@ static int get_tmpname(char *fnametmp, char *fname)
 	maxname = MIN(MAXPATHLEN - 7 - length, NAME_MAX - 8);
 
 	if (maxname < 1) {
-		rprintf(FERROR, "temporary filename too long: %s\n",
-			safe_fname(fname));
+		rprintf(FERROR, "temporary filename too long: %s\n", fname);
 		fnametmp[0] = '\0';
 		return 0;
 	}
@@ -204,13 +140,39 @@ static int receive_data(int f_in, char *fname_r, int fd_r, OFF_T size_r,
 		mapbuf = map_file(fd_r, size_r, read_size, sum.blength);
 		if (verbose > 2) {
 			rprintf(FINFO, "recv mapped %s of size %.0f\n",
-				safe_fname(fname_r), (double)size_r);
+				fname_r, (double)size_r);
 		}
 	} else
 		mapbuf = NULL;
 
 	sum_init(checksum_seed);
 
+	if (append_mode) {
+		OFF_T j;
+		sum.flength = (OFF_T)sum.count * sum.blength;
+		if (sum.remainder)
+			sum.flength -= sum.blength - sum.remainder;
+		for (j = CHUNK_SIZE; j < sum.flength; j += CHUNK_SIZE) {
+			if (do_progress)
+				show_progress(offset, total_size);
+			sum_update(map_ptr(mapbuf, offset, CHUNK_SIZE),
+				   CHUNK_SIZE);
+			offset = j;
+		}
+		if (offset < sum.flength) {
+			int32 len = sum.flength - offset;
+			if (do_progress)
+				show_progress(offset, total_size);
+			sum_update(map_ptr(mapbuf, offset, len), len);
+			offset = sum.flength;
+		}
+		if (fd != -1 && do_lseek(fd, offset, SEEK_SET) != offset) {
+			rsyserr(FERROR, errno, "lseek failed on %s",
+				full_fname(fname));
+			exit_cleanup(RERR_FILEIO);
+		}
+	}
+
 	while ((i = recv_token(f_in, &data)) != 0) {
 		if (do_progress)
 			show_progress(offset, total_size);
@@ -314,22 +276,22 @@ static void handle_delayed_updates(struct file_list *flist, char *local_name)
 	char *fname, *partialptr, numbuf[4];
 	int i;
 
-	for (i = -1; (i = next_delayed_bit(i)) >= 0; ) {
+	for (i = -1; (i = bitbag_next_bit(delayed_bits, i)) >= 0; ) {
 		struct file_struct *file = flist->files[i];
-		fname = local_name ? local_name : f_name(file);
+		fname = local_name ? local_name : f_name(file, NULL);
 		if ((partialptr = partial_dir_fname(fname)) != NULL) {
 			if (make_backups && !make_backup(fname))
 				continue;
 			if (verbose > 2) {
 				rprintf(FINFO, "renaming %s to %s\n",
-					safe_fname(partialptr),
-					safe_fname(fname));
+					partialptr, fname);
 			}
+			/* We don't use robust_rename() here because the
+			 * partial-dir must be on the same drive. */
 			if (do_rename(partialptr, fname) < 0) {
 				rsyserr(FERROR, errno,
 					"rename failed for %s (from %s)",
-					full_fname(fname),
-					safe_fname(partialptr));
+					full_fname(fname), partialptr);
 			} else {
 				if (remove_sent_files
 				    || (preserve_hard_links
@@ -337,8 +299,7 @@ static void handle_delayed_updates(struct file_list *flist, char *local_name)
 					SIVAL(numbuf, 0, i);
 					send_msg(MSG_SUCCESS,numbuf,4);
 				}
-				handle_partial_dir(partialptr,
-						   PDIR_DELETE);
+				handle_partial_dir(partialptr, PDIR_DELETE);
 			}
 		}
 	}
@@ -351,7 +312,7 @@ static int get_next_gen_i(int batch_gen_fd, int next_gen_i, int desired_i)
 			rprintf(FINFO,
 				"(No batched update for%s \"%s\")\n",
 				phase ? " resend of" : "",
-				safe_fname(f_name(the_file_list->files[next_gen_i])));
+				f_name(the_file_list->files[next_gen_i], NULL));
 		}
 		next_gen_i = read_int(batch_gen_fd);
 		if (next_gen_i == -1)
@@ -394,7 +355,7 @@ int recv_files(int f_in, struct file_list *flist, char *local_name)
 	}
 
 	if (delay_updates)
-		init_delayed_bits(flist->count);
+		delayed_bits = bitbag_create(flist->count);
 
 	while (1) {
 		cleanup_disable();
@@ -416,6 +377,10 @@ int recv_files(int f_in, struct file_list *flist, char *local_name)
 			send_msg(MSG_DONE, "", 0);
 			if (keep_partial && !partial_dir)
 				make_backups = 0; /* prevents double backup */
+			if (append_mode) {
+				append_mode = 0;
+				sparse_files = 0;
+			}
 			continue;
 		}
 
@@ -425,10 +390,10 @@ int recv_files(int f_in, struct file_list *flist, char *local_name)
 			continue;
 
 		file = flist->files[i];
-		fname = local_name ? local_name : f_name_to(file, fbuf);
+		fname = local_name ? local_name : f_name(file, fbuf);
 
 		if (verbose > 2)
-			rprintf(FINFO, "recv_files(%s)\n", safe_fname(fname));
+			rprintf(FINFO, "recv_files(%s)\n", fname);
 
 		if (!(iflags & ITEM_TRANSFER)) {
 			maybe_log_item(file, iflags, itemizing, xname);
@@ -452,19 +417,26 @@ int recv_files(int f_in, struct file_list *flist, char *local_name)
 			exit_cleanup(RERR_PROTOCOL);
 		}
 
-		if (dry_run) { /* log the transfer */
+		if (!do_xfers) { /* log the transfer */
 			if (!am_server && log_format)
 				log_item(file, &stats, iflags, NULL);
 			if (read_batch)
 				discard_receive_data(f_in, file->length);
 			continue;
 		}
+		if (write_batch < 0) {
+			log_item(file, &stats, iflags, NULL);
+			if (!am_server)
+				discard_receive_data(f_in, file->length);
+			continue;
+		}
 
 		if (read_batch) {
 			next_gen_i = get_next_gen_i(batch_gen_fd, next_gen_i, i);
 			if (i < next_gen_i) {
-				rprintf(FINFO, "(Skipping batched update for \"%s\")\n",
-					safe_fname(fname));
+				rprintf(FINFO,
+					"(Skipping batched update for \"%s\")\n",
+					fname);
 				discard_receive_data(f_in, file->length);
 				continue;
 			}
@@ -565,16 +537,16 @@ int recv_files(int f_in, struct file_list *flist, char *local_name)
 			fd1 = -1;
 		}
 
-		if (fd1 != -1 && !preserve_perms) {
-			/* if the file exists already and we aren't preserving
-			 * permissions then act as though the remote end sent
-			 * us the file permissions we already have */
-			file->mode = st.st_mode;
+		/* If we're not preserving permissions, change the file-list's
+		 * mode based on the local permissions and some heuristics. */
+		if (!preserve_perms) {
+			int exists = fd1 != -1;
+			file->mode = dest_mode(file->mode, st.st_mode, exists);
 		}
 
 		/* We now check to see if we are writing file "inplace" */
 		if (inplace)  {
-			fd2 = do_open(fname, O_WRONLY|O_CREAT, 0);
+			fd2 = do_open(fname, O_WRONLY|O_CREAT, 0600);
 			if (fd2 == -1) {
 				rsyserr(FERROR, errno, "open %s failed",
 					full_fname(fname));
@@ -603,7 +575,7 @@ int recv_files(int f_in, struct file_list *flist, char *local_name)
 			 * because their information should have been previously
 			 * transferred, but that may not be the case with -R */
 			if (fd2 == -1 && relative_paths && errno == ENOENT
-			    && create_directory_path(fnametmp, orig_umask) == 0) {
+			    && create_directory_path(fnametmp) == 0) {
 				/* Get back to name with XXXXXX in it. */
 				get_tmpname(fnametmp, fname);
 				fd2 = do_mkstemp(fnametmp, file->mode & INITACCESSPERMS);
@@ -617,7 +589,7 @@ int recv_files(int f_in, struct file_list *flist, char *local_name)
 				continue;
 			}
 
-			if (partialptr)
+			if (keep_partial)
 				cleanup_set(fnametmp, partialptr, file, fd1, fd2);
 		}
 
@@ -625,7 +597,7 @@ int recv_files(int f_in, struct file_list *flist, char *local_name)
 		if (log_before_transfer)
 			log_item(file, &initial_stats, iflags, NULL);
 		else if (!am_server && verbose && do_progress)
-			rprintf(FINFO, "%s\n", safe_fname(fname));
+			rprintf(FINFO, "%s\n", fname);
 
 		/* recv file data */
 		recv_ok = receive_data(f_in, fnamecmp, fd1, st.st_size,
@@ -643,17 +615,20 @@ int recv_files(int f_in, struct file_list *flist, char *local_name)
 		}
 
 		if ((recv_ok && (!delay_updates || !partialptr)) || inplace) {
-			finish_transfer(fname, fnametmp, file, recv_ok, 1);
-			if (partialptr != fname && fnamecmp == partialptr) {
+			if (partialptr == fname || *partial_dir == '/')
+				partialptr = NULL;
+			finish_transfer(fname, fnametmp, partialptr,
+					file, recv_ok, 1);
+			if (fnamecmp == partialptr) {
 				do_unlink(partialptr);
 				handle_partial_dir(partialptr, PDIR_DELETE);
 			}
 		} else if (keep_partial && partialptr
 		    && handle_partial_dir(partialptr, PDIR_CREATE)) {
-			finish_transfer(partialptr, fnametmp, file, recv_ok,
-					!partial_dir);
+			finish_transfer(partialptr, fnametmp, NULL,
+					file, recv_ok, !partial_dir);
 			if (delay_updates && recv_ok) {
-				set_delayed_bit(i);
+				bitbag_set_bit(delayed_bits, i);
 				recv_ok = -1;
 			}
 		} else {
@@ -688,8 +663,7 @@ int recv_files(int f_in, struct file_list *flist, char *local_name)
 				}
 				rprintf(msgtype,
 					"%s: %s failed verification -- update %s%s.\n",
-					errstr, safe_fname(fname),
-					keptstr, redostr);
+					errstr, fname, keptstr, redostr);
 			}
 			if (!phase) {
 				SIVAL(numbuf, 0, i);
