@@ -28,6 +28,9 @@
 
 int modify_window = 0;
 int module_id = -1;
+int relative_paths = 0;
+int human_readable = 0;
+mode_t orig_umask = 002;
 char *partial_dir;
 struct filter_list_struct server_filter_list;
 
